#
# Copyright (C) 2022 Soleta Networks <info@soleta.eu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

import os

import fdisk


def get_disks():
    """
    Walks /sys/block/ and returns files starting with 'sd',
    'nvme' or 'vd'
    """
    return sorted([ dev for dev in os.listdir('/sys/block/')
                    if dev.startswith('sd')
                    or dev.startswith('nvme')
                    or dev.startswith('vd')])


def get_partition_device(disknum, partnum):
    """
    Returns the device path, given a disk and partition number
    """
    disk = get_disks()[disknum-1]
    cxt = fdisk.Context(f'/dev/{disk}')

    for pa in cxt.partitions:
        if pa.partno == partnum - 1:
            return cxt.partition_to_string(pa, fdisk.FDISK_FIELD_DEVICE)

    raise ValueError('No such partition')


def get_efi_partition(disknum):
    """
    Look for an EFI System Partition at the n-th disk. If disknum is invalid an exception is thrown

    Returns tuple with:
        - Device name containing the ESP
        - /dev/{device} string
        - Partition number (starting at 1)
    """
    try:
        disk = get_disks()[disknum-1]
        cxt = fdisk.Context(f'/dev/{disk}')

        if cxt.label == fdisk.FDISK_DISKLABEL_DOS:
            raise RuntimeError('Disk has DOS partition scheme, cannot find ESP.')

        for pa in cxt.partitions:
            if pa.type.name == 'EFI System':
                return cxt.partition_to_string(pa, fdisk.FDISK_FIELD_DEVICE), f'/dev/{disk}', pa.partno + 1
    except:
        logging.error(f'Unable to find efi partition at disk number {disknum}')
        raise
